package cs214.webapp
package client

import java.net.{URLDecoder, URLEncoder}
import scala.concurrent.ExecutionContext.Implicits.global
import org.scalajs.dom
import org.scalajs.dom.{Element, KeyCode}
import org.scalajs.dom.html.{Input, Select, TextArea}
import scalatags.JsDom.all.*

/** Parent class to web pages */
abstract class Page:
  val classList: String
  def renderInto(target: dom.Element): Unit

  def pageHeader(subtitle: String) =
    frag(h1(cls := "title", "ScalApp"), h2(subtitle))

  def replaceChildren(target: dom.Element)(frag: Frag): Unit =
    target.replaceChildren(frag.render)
    target.setAttribute("class", classList)

  private def path: Seq[String] =
    this match
      case HomePage =>
        Nil
      case InstanceCreationPage(appID) =>
        List("app", appID)
      case UIPage(appID, instanceID) => 
        List("app", appID, instanceID)
      case JoinPageLoader(appID, uiID, instanceID) =>
        List("app", appID, uiID, instanceID)
      case AppPage(appID, uiID, instanceID, userID) =>
        List("app", appID, uiID, instanceID, userID)
      case _ =>
        throw IllegalArgumentException(f"No path for $this!")

  def url =
    f"/${path.map(s => URLEncoder.encode(s, "UTF-8")).mkString("/")}"

object Page:
  /** Creates a component from a URL path
    * @param appInfos
    *   retrieves the app information given its id
    */
  def from(url: String): Page =
    val components = url.stripPrefix("/").stripSuffix("/").split("/")
    val decoded = components.map(c => URLDecoder.decode(c, "UTF-8")).toList
    decoded match
      case List("") =>
        HomePage
      case List("app", appID) =>
        InstanceCreationPage(appID)
      case List("app", appID, uiID) =>
        UIPage(appID, uiID)
      case List("app", appID, uiID, instanceID) =>
        JoinPageLoader(appID, uiID, instanceID)
      case List("app", appID, uiID, instanceID, userID) =>
        AppPage(appID, uiID, instanceID, userID)
      case _ =>
        throw IllegalArgumentException(f"Unknown url $url!")

/** The initial home page which fetches the list of apps */
object HomePage extends Page:
  val classList = "HomePage"

  def renderInto(target: Element) =
    Requests.listApps.map: response =>
      AppSelectionPage(response.apps).renderInto(target)

/** The app selection menu, where the user can create a new app. */
case class AppSelectionPage(apps: Seq[AppInfo]) extends Page:
  val classList = "AppSelectionPage"

  def selectApp(appID: AppID): Unit =
    WebClient.navigateTo(InstanceCreationPage(appID))

  def renderInto(target: Element) = replaceChildren(target):
    frag(
      pageHeader("Select an app"),
      div(
        id := "app-list",
        apps.groupBy(_.year).toList.map: (year, appsOfYear) =>
          frag(
            h5(year.toString),
            div(
              cls := "app-grid", {
              appsOfYear.sortBy(_.name).map: appInfo =>
                div(
                  cls := "app-option",
                  onclick := (() => selectApp(appInfo.id)),
                  img(src := s"./static/${appInfo.id}.png", cls := "app-thumbnail"),
                  p(cls := "app-option-name", appInfo.name),
                  p(cls := "app-option-description", appInfo.description)
                )
              }
            )
          )
      )
    )

/** The instance creation menu, where you enter the player names */
case class InstanceCreationPage(appID: AppID) extends Page:
  val classList = "InstanceCreationPage"

  def submit(e: dom.Event): Unit =
    e.preventDefault()
    val userIDstr = getElementById[Input]("user-IDs").value
    val userIDs = userIDstr.split("[;,]").map(_.strip).to(Seq).distinct
    Requests.createInstance(appID, userIDs).map: resp =>
      WebClient.navigateTo(UIPage(appID, resp.instanceID))

  def renderInto(target: Element) = replaceChildren(target):
    frag(
      pageHeader(s"Create a new $appID instance"),
      form(
        onsubmit := submit,
        div(
          cls := "grid-form",
          label(`for` := "user-ids", "User IDs: "),
          input(
            `type` := "text",
            id := "user-ids",
            placeholder := "user1; user2; …",
            required := true,
            autofocus := true
          )
        ),
        input(`type` := "submit", value := "Start!")
      )
    )

/** The ui choice menu */
case class UIPage(appID: AppID, instanceID: InstanceID) extends Page:
  val classList = "UIPage"

  private val appUIs = WebClient.getApps(appID)

  private def cssID(idx: Int) = f"ui-$idx"

  private def handleKeyboardEvent(e: dom.KeyboardEvent): Unit =
    if e.keyCode == KeyCode.Enter then
      e.preventDefault()
      dom.window.removeEventListener("keydown", (e: dom.KeyboardEvent) => handleKeyboardEvent(e))
      joinPageLoader()

  private def handleFormSubmission(e: dom.Event): Unit =
    e.preventDefault()
    joinPageLoader()

  private def getSelected =
    appUIs.zipWithIndex.find { (ui, i) =>
      getElementById[Input](cssID(i)).checked
    }.map(_._1)

  private def joinPageLoader(): Unit =
    getSelected.map: ui =>
      WebClient.navigateTo(JoinPageLoader(appID, ui.uiID, instanceID))

  def renderInto(target: Element) = 
    if appUIs.size <= 1 then
      WebClient.navigateTo(JoinPageLoader(appID, appUIs(0).uiID, instanceID))
    else replaceChildren(target):
      dom.window.addEventListener("keydown", (e: dom.KeyboardEvent) => handleKeyboardEvent(e))  
      frag(
        pageHeader("Choose your UI"),
        form(
          onsubmit := handleFormSubmission,
          fieldset(
            legend("Select your UI"),
            for (appUI, idx) <- appUIs.zipWithIndex
            yield div(
              input(
                `type` := "radio",
                id := cssID(idx),
                if idx == 0 then checked := "checked" else frag(),
                name := "ui",
                value := appUI.uiID,
                required := true
              ),
              label(`for` := cssID(idx), appUI.uiID)
            )
          ),
          input(`type` := "submit", value := "Play!")
        )
      )

/** The pre-connection menu, which fetches the user list. */
case class JoinPageLoader(appID: AppID, uiID: UIID, instanceID: InstanceID) extends Page:
  val classList = "JoinPageLoader"
  def renderInto(target: Element) =
    Requests.instanceInfo(instanceID).map: resp =>
      JoinPage(appID, uiID, instanceID, resp.userIDs).renderInto(target)

/** The connection menu, where a user joins an existing app. */
case class JoinPage(appID: AppID, uiID: UIID, instanceID: InstanceID, userIDs: Seq[UserID])
    extends Page:
  val classList = "JoinPage"

  private def cssID(idx: Int) = f"user-$idx"

  private def handleKeyboardEvent(e: dom.KeyboardEvent): Unit =
    if e.keyCode == KeyCode.Enter then
      e.preventDefault()
      dom.window.removeEventListener("keydown", (e: dom.KeyboardEvent) => handleKeyboardEvent(e))
      joinAppInstance()

  private def handleFormSubmission(e: dom.Event): Unit =
    e.preventDefault()
    joinAppInstance()

  private def getSelected =
    userIDs.zipWithIndex.find { (u, i) =>
      getElementById[Input](cssID(i)).checked
    }.map(_._1)

  private def joinAppInstance(): Unit =
    getSelected.map: userID =>
      WebClient.navigateTo(AppPage(appID, uiID, instanceID, userID))

  def renderInto(target: Element): Unit = replaceChildren(target):
    dom.window.addEventListener("keydown", (e: dom.KeyboardEvent) => handleKeyboardEvent(e))
    frag(
      pageHeader("Join app instance"),
      form(
        onsubmit := handleFormSubmission,
        fieldset(
          legend("Select your username"),
          for (userID, idx) <- userIDs.zipWithIndex
          yield div(
            input(
              `type` := "radio",
              id := cssID(idx),
              if idx == 0 then checked := "checked" else frag(),
              name := "user",
              value := userID,
              required := true
            ),
            label(`for` := cssID(idx), userID)
          )
        ),
        input(`type` := "submit", value := "Join!")
      )
    )

/** The actual app content page */
case class AppPage(appID: AppID, uiID: UIID, instanceID: InstanceID, userID: UserID) extends Page:
  val classList = f"app $appID"

  private val app: ClientApp =
    val apps = WebClient.getApps(appID)
    require(
      !apps.isEmpty,
      s"Could not find an app UI corresponding to app ID \"$appID\". There may be a " +
        s"mismatch between the client UI's appID and server logic's appID, or you may have " +
        s"forgotten to annotate your UI with @JSExportTopLevel(\"<yourAppID>\")"
    )
    val maybeApp = apps.find(_.uiID == uiID)
    require(
      maybeApp.isDefined,
      s"Could not find an app UI with ID \"$uiID\"."
    )
    maybeApp.get

  def renderInto(target: Element) =
    replaceChildren(target):
      frag(header(id := "banner"), tag("section")(id := "app"))
    Requests.instanceInfo(instanceID).map: appInfo =>
      val hostName = dom.window.location.hostname
      val endpoint = appInfo.wsEndpoint
        .replace("{{hostName}}", hostName)
        .replace("{{userID}}", URLEncoder.encode(userID, "UTF-8"))
      IpBanner(appInfo.shareUrl).renderInto(target.querySelector("#banner"))
      app.init(instanceID, userID, endpoint, target.querySelector("#app"))

/** The top-most share banner */
case class IpBanner(shareUrl: String) extends Page:
  val classList = "IpBanner"

  def renderInto(target: Element) =
    replaceChildren(target):
      val copyToClipboardButtonText = "📋 Copy to clipboard."
      frag(
        span("Share ", a(href := shareUrl, shareUrl), " with your friends to let them join!"),
        div(
          id := "banner-actions",
          button(
            onclick := (() => WebClient.navigateTo(HomePage)),
            "🏠 Back to home page"
          ),
          button(
            id := "copy-share-url",
            onclick := (() =>
              dom.window.navigator.clipboard.writeText(shareUrl)
              dom.document.getElementById("copy-share-url").innerText = "Copied!"

              dom.window.setTimeout(
                () => dom.document.getElementById("copy-share-url").innerText = copyToClipboardButtonText,
                1000
              )
            ),
            copyToClipboardButtonText
          )
        )
      )

def getElementById[T](id: String) =
  dom.document.getElementById(id).asInstanceOf[T]
