package cs214.webapp
package client

import scala.collection.mutable
import scala.scalajs.js.timers.{SetTimeoutHandle, setTimeout}
import scala.util.{Failure, Success}
import org.scalajs.dom
import scalatags.JsDom.all.Frag

type Target = dom.Element

abstract class WSClientApp extends ClientApp:
  WebClient.register(this)

  protected def init(userID: UserID, sendMessage: ujson.Value => Unit, target: Target): ClientAppInstance

  def init(instanceID: InstanceID, userID: UserID, endpoint: String, target: Target): ClientAppInstance =
    val socket = new dom.WebSocket(endpoint)
    socket.onopen = (event: dom.Event) => println("WebSocket connection opened")
    socket.onclose = (event: dom.CloseEvent) => println(s"WebSocket connection closed: ${event.reason}")
    socket.onerror = (event: dom.Event) => println(s"WebSocket error: ${event.`type`}")

    val sendMessage = (js: ujson.Value) => socket.send(js.toString)
    val client = init(userID, sendMessage, target)
    socket.onmessage = msg =>
      println(msg)
      val js = ujson.read(msg.data.toString)
      client.onMessage(SocketResponseWire.decode(js).flatten)

    client

/** Instance of a client-side state machine application.
  *
  * @param userID
  *   the user for which creating the client application
  * @param sendMessage
  *   how to send a json message
  * @param target
  *   the element where to draw the application
  */
abstract class StateMachineClientAppInstance[Event, View](
    userID: UserID,
    sendMessage: ujson.Value => Unit,
    target: Target
) extends ClientAppInstance:

  /** Optional CSS style sheet to apply to the application page. When
    * overriding, do not forget to append to the parent definition.
    */
  def css: String = ""
  if !css.isEmpty then
    // Add <style> tag in the header of the page
    val styleTag = dom.document.createElement("style")
    styleTag.textContent = css
    dom.document.head.appendChild(styleTag)

  /** Provides serialization methods for events and views */
  def wire: AppWire[Event, View]

  /** Renders a [[View]] received from the server. The method also takes a
    * [[UserID]] to get information on the context and an [[onEvent]] callback
    * which it uses to send server events when specific actions are triggered on
    * the UI (click on a button for example)
    * @param userID
    *   The user ID used by the client.
    * @param view
    *   The view to render.
    * @return
    *   The rendered view.
    */
  def render(userID: UserID, view: View): Frag

  /** Renders a non-fatal error as a browser alert */
  private def renderError(msg: String): Unit =
    dom.window.alert(msg)

  protected def sendEvent(event: Event) =
    sendMessage(wire.eventFormat.encode(event))

  /** @inheritdoc */
  override def onMessage(msg: util.Try[ujson.Value]): Unit =
    msg.flatMap(EventResponse.Wire.decode) match
      case Failure(msg)              => WebClient.crash(Exception(msg))
      case Success(Failure(msg))     => renderError(msg.getMessage)
      case Success(Success(actions)) => actionsQueue.enqueueAll(actions)
    tryProcessNextAction()

  /** The views to render */
  private val actionsQueue: mutable.Queue[Action[ujson.Value]] = mutable.Queue()

  /** Minimal cooldown time between the render of two views */
  private var actionCooldownDelay = 0

  /** The timestamp for the last render of a view */
  private var lastActionAppliedMs = System.currentTimeMillis()

  /** A call to the [[tryProcessNextAction()]] function scheduled */
  private var timeout: Option[SetTimeoutHandle] = None

  /** Sets the [[actionCooldownDelay]] of the client to the specified value */
  private def setCooldown(ms: Int): Unit = actionCooldownDelay = ms

  /** Apply the next action from [[actionsQueue]] if [[actionCooldownDelay]]
    * permits it.
    */
  private def tryProcessNextAction(): Unit =
    // If there are still views to render and if no call to this function is scheduled
    if actionsQueue.nonEmpty && timeout.isEmpty then
      // Then check if we are out of the cooldown delay
      if System.currentTimeMillis() - lastActionAppliedMs > actionCooldownDelay then
        // We can render the view, and dequeue it
        processAction(actionsQueue.dequeue())
        // Continue try emptying the queue
        tryProcessNextAction()
      else
        // We still have to wait, put a timeout to call the function later
        timeout = Some(setTimeout(actionCooldownDelay) {
          // First remove the timeout so that, if necessary,
          // the next function call can create a new one
          timeout = None
          // Then try to render next views
          tryProcessNextAction()
        })

  /** Execute a single action sent by the server. */
  private def processAction(jsonAction: Action[ujson.Value]): Unit =
    lastActionAppliedMs = System.currentTimeMillis()
    setCooldown(0)
    jsonAction match
      case Action.Alert(msg) =>
        dom.window.alert(msg)
      case Action.Pause(durationMs) =>
        setCooldown(durationMs)
      case Action.Render(js) =>
        // Step1: The client receives the view sent by the server here
        wire.viewFormat.decode(js) match
          case Failure(exception) => renderError(exception.getMessage)
          case Success(jsonView) =>
            target.replaceChildren:
              this.render(userID, jsonView).render
