package cs214.webapp
package client

import scala.collection.mutable
import scala.util.{Failure, Success}
import org.scalajs.dom

object WebClient:
  private val appLibrary: mutable.Map[AppID, Seq[ClientApp]] = mutable.Map.empty

  val getApps = (appID: AppID) => appLibrary.getOrElse(appID, Seq.empty[ClientApp])

  def register(clientApp: ClientApp): Unit =
    println(s"Registered ${clientApp.appID}/${clientApp.uiID}'s UI")
    val apps = appLibrary.getOrElse(clientApp.appID, Nil)
    appLibrary.put(clientApp.appID, apps :+ clientApp)

  def start(): Unit =
    start(dom.document.getElementById("root"))

  def start(root: dom.Element): Unit =
    println(s"Registered apps: ${appLibrary.keys.toSeq.sorted}.")
    try
      Page.from(dom.document.location.pathname)
        .renderInto(root)
    catch case t => crash(t)

  def navigateTo(page: Page) =
    dom.window.location.pathname = page.url

  def crash(t: Throwable) =
    dom.document.querySelector("body").prepend:
      import scalatags.JsDom.all.*
      tag("dialog")(id := "fatal", b("Fatal error: "), t.getMessage).render
    dom.document.querySelector("#fatal").asInstanceOf[dom.html.Dialog].showModal()
    throw t

object Requests:
  import org.scalajs.dom.{Fetch, Headers}

  import concurrent.ExecutionContext.Implicits.global
  import scala.concurrent.Future
  import scala.scalajs.js.Promise

  /** Sends a POST request to create a new instance including the given users of
    * the app with the given id
    */
  def createInstance(appID: AppID, userIDs: Seq[UserID]): Future[CreateInstanceResponse] =
    sendPostRequestWith(Endpoints.Api.createInstance)(CreateInstanceRequest.Wire, CreateInstanceResponse.Wire):
      CreateInstanceRequest(appID, userIDs)

  /** Sends a GET request to enumerate available apps */
  def listApps: Future[ListAppsResponse] =
    Fetch.fetch(Endpoints.Api.listApps.toString)
      .decodeResponse(ListAppsResponse.Wire)
      .logErrors

  /** Sends a GET request to retrieve information on an app. */
  def instanceInfo(instanceID: InstanceID): Future[InstanceInfoResponse] =
    Fetch.fetch(f"${Endpoints.Api.instanceInfo.toString}?instanceID=$instanceID")
      .decodeResponse(InstanceInfoResponse.Wire)
      .logErrors

  /** Sends a POST request with the given payload to the specified endpoint */
  private def sendPostRequestWith[R, S](endpoint: Endpoints.Api)(
      reqWire: WireFormat[R],
      respWire: WireFormat[S]
  )(payload: R) =
    val requestHeaders = dom.Headers()
    requestHeaders.append("Content-Type", "application/json")

    val requestOptions = new dom.RequestInit:
      method = dom.HttpMethod.POST
      headers = requestHeaders
      body = reqWire.encode(payload).toString

    Fetch.fetch(endpoint.toString, requestOptions)
      .decodeResponse(respWire)
      .logErrors

  extension [T](p: Future[T])
    private def logErrors =
      p.andThen { case Failure(t) => WebClient.crash(t) }

  case class FailedRequestException(status: Int, msg: String) extends Exception(msg)

  extension (p: Promise[dom.Response])
    private def decodeResponse[R](wire: WireFormat[R]) =
      p.toFuture.flatMap: response =>
        response.text().toFuture.flatMap: txt =>
          if !response.ok then
            throw FailedRequestException(response.status, txt)
          Future.fromTry(wire.decode(ujson.read(txt)))
