package cs214.webapp.client
package graphics

import cs214.webapp.UserID
import cs214.webapp.client.StateMachineClientAppInstance
import org.scalajs.dom
import org.scalajs.dom.HTMLInputElement
import scalatags.JsDom
import scalatags.JsDom.all.*

enum HTMLAttribute:
  case cls
  case style

type CSSProperties = Map[String, String]
extension (cssProps: CSSProperties)
  def toCSS(): String = cssProps
    .map((k, v) => f"${k}: ${v}")
    .mkString("; ")

enum MouseEvent:
  case Click(button: MouseButton)
  case MouseDown(button: MouseButton)
  case MouseUp(button: MouseButton)
  case HoverEnter
  case HoverLeave

/** The console is segmented with [[TextSegment]]s. Each segment represents a
  * piece of text with its own modifiers and event listeners.
  *
  * If multiple parts of your text need to respond independently to mouse
  * events or need a different style, you should consider separating them in
  * different [[TextSegment]]s
  *
  * @param text
  *   The text to be displayed
  * @param onMouseEvent
  *   How to handle mouse events happening on this text segment
  * @param cssProperties
  *   The css properties to inline with this text segment (html style
  *   attribute)
  * @param htmlAttributes
  *   The html attributes to add to this text segment
  * @param modifiers
  *   Any additional `JsDom.Modifier`s you may want to add to the element
  */
case class TextSegment(
    text: String,
    onMouseEvent: Option[MouseEvent => Unit] = None,
    cssProperties: CSSProperties = Map(),
    htmlAttributes: Map[HTMLAttribute, String] = Map(),
    modifiers: scalatags.JsDom.Modifier*
)

/** Text-based UI. Detects mouse events and has an integrated text field */
abstract class TextClientAppInstance[Event, View](
    userID: UserID,
    sendMessage: ujson.Value => Unit,
    target: Target
) extends StateMachineClientAppInstance[Event, View](userID, sendMessage, target):
  /** Transforms a text input from the user into an optional app [[Event]]
    * @param view
    *   The current view
    * @param text
    *   The input text typed by the user
    * @return
    *   An optional application event
    */
  def handleTextInput(view: View, text: String): Option[Event]

  /** Transforms a view received by the server into a segmented console display.
    * Each [[TextSegment]] responds independently to mouse events and defines
    * its own style.
    *
    * @param userID
    *   The ID of the user using this client
    * @param view
    *   The view of the application received by the server
    * @return
    *   The different text segments of the console
    */
  def renderView(userID: UserID, view: View): Vector[TextSegment]

  override def render(userID: UserID, view: View): JsDom.all.Frag =
    currView = Some(view)

    frag(
      div(
        cls := "textapp",
        div(
          cls := "console",
          pre(
            style := "white-space: pre-wrap;",
            for textSegment <- renderView(userID, view)
            yield span(
              textSegment.text,
              style := textSegment.htmlAttributes.getOrElse(HTMLAttribute.style, textSegment.cssProperties.toCSS()),
              cls := textSegment.htmlAttributes.getOrElse(HTMLAttribute.cls, ""),
              textSegment.modifiers,
              margin := "0",
              height := "1em",
              onclick := {
                (e: dom.MouseEvent) =>
                  if textSegment.onMouseEvent.isDefined then
                    textSegment.onMouseEvent.get(MouseEvent.Click(MouseButton.from(e.button)))
              },
              onmousedown := {
                (e: dom.MouseEvent) => 
                  if textSegment.onMouseEvent.isDefined then
                    textSegment.onMouseEvent.get(MouseEvent.MouseDown(MouseButton.from(e.button)))
              },
              onmouseup := { (e: dom.MouseEvent) =>
                if textSegment.onMouseEvent.isDefined then
                  textSegment.onMouseEvent.get(MouseEvent.MouseUp(MouseButton.from(e.button)))
              },
              onmouseover := { (_: dom.MouseEvent) => 
                if textSegment.onMouseEvent.isDefined then
                  textSegment.onMouseEvent.get(MouseEvent.HoverEnter) },
              onmouseout := { (_: dom.MouseEvent) => 
                if textSegment.onMouseEvent.isDefined then
                  textSegment.onMouseEvent.get(MouseEvent.HoverLeave) }
            )
          )
        ),
        input(
          `type` := "text",
          autofocus := true,
          id := "console-input",
          tabindex := 0,
          onkeydown := handleKeyDown,
        )
      )
    )

  private var currView: Option[View] = None

  protected def focusConsoleInput(): Unit =
    dom.document.getElementById("console-input").asInstanceOf[HTMLInputElement].focus()

  private def handleKeyDown(e: dom.KeyboardEvent): Unit = currView
    .flatMap(view =>
      if e.keyCode == dom.KeyCode.Enter then
        e.target match
          case inputElement: HTMLInputElement =>
            e.preventDefault()
            val v = inputElement.value
            inputElement.value = ""
            handleTextInput(view, v)
      else
        None
    ).foreach(sendEvent)

  override def css: String = super.css +
    """
    | .textapp {
    |   font-size: 1rem;
    | }
    |
    | .textapp input {
    |   height: 1em;
    |   width: 100%;
    |   border: solid 1px black;
    |   box-sizing: border-box;
    | }
    |  
    |  .textapp .console {
    |      overflow: hidden;
    |      padding: 0.5em;
    |      font-size: 1em;
    |      display: flex;
    |      flex-direction: column;
    |      justify-content: center;
    |      align-items: start;
    |      border: solid 1px black;
    |  }
    |  
    |  .textapp .console span {
    |      box-sizing: border-box;
    |  }
    """.stripMargin
