package cs214.webapp
package server
package web

import scala.util.Try

/** Server-side apps definition and abstractions */

/** A server-side app, allows creating new `ServerAppInstance` */
private[web] sealed trait ServerApp:
  def appInfo: AppInfo
  def init(clients: Seq[UserID]): ServerAppInstance

/** A server-side app instance */
private[web] sealed trait ServerAppInstance:
  val appInfo: AppInfo
  val registeredUsers: Seq[String]

  /** Compute setup actions to send to client. */
  def respondToNewClient(userID: UserID): ujson.Value

  /** Simulate one step of the underlying state machine. */
  def transition(clients: Seq[UserID])(
      userID: UserID,
      jsEvent: ujson.Value
  ): Try[(Map[UserID, Seq[Action[ujson.Value]]], ServerAppInstance)]

/** Wraps a state machine into a server-side app. */
private[web] case class StateMachineServerApp[E, S, V](sm: StateMachine[E, S, V]) extends ServerApp:
  def appInfo: AppInfo = sm.appInfo

  def init(clients: Seq[UserID]): ServerAppInstance =
    StateMachineServerAppInstance(sm = sm, registeredUsers = clients, state = sm.init(clients))

/** Wraps a state machine and its state into a server-side app instance. */
private[web] case class StateMachineServerAppInstance[E, S, V](
    sm: StateMachine[E, S, V],
    registeredUsers: Seq[UserID],
    state: S
) extends ServerAppInstance:

  val appInfo: AppInfo = sm.appInfo

  def respondToNewClient(userID: UserID): ujson.Value =
    sm.wire.viewFormat.encode(sm.project(state)(userID))

  def transition(clients: Seq[UserID])(
      uID: UserID,
      eventJs: ujson.Value
  ): Try[(Map[UserID, Seq[Action[ujson.Value]]], ServerAppInstance)] = Try:
    val event = sm.wire.eventFormat.decode(eventJs).get
    if WebServer.debug then
      println(f"[debug] got event: $event")

    val actions = sm.transition(state)(uID, event).get
    if WebServer.debug then
      println(s"[debug] transition produced actions ${actions.mkString(", ")}")

    val clientActions = clients.map: uID =>
      uID -> projectFor(uID)(actions).map(_.map(sm.wire.viewFormat.encode))
    val nextState =
      actions.collect { case Action.Render(t) => t }.lastOption.getOrElse(state)

    (clientActions.toMap, copy(state = nextState))

  private def projectFor(userID: UserID)(actions: Seq[Action[S]]): Seq[Action[V]] =
    actions.map(_.map(st => sm.project(st)(userID)))
