package cs214.webapp
package server
package web

import java.util.concurrent.atomic.AtomicBoolean
import scala.collection.immutable.Map
import scala.collection.{concurrent, immutable, mutable}
import scala.util.{Failure, Success, Try}
import scala.io.Source
import scala.io.Codec

/** Contains the web server state and functionalities */
object WebServer:
  private val instanceIDsFileName = "instance_ids" // Where to found the list of possible instance IDs
  private val instanceIDsCount = 4 // How many words do we concatenate for the instance id
  
  private[web] val debug = false

  private[web] case class RunningClock(
      clockID: UserID,
      running: AtomicBoolean
  )

  private[web] case class RunningServerAppInstance(
      instance: ServerAppInstance,
      connectedUsersCount: Int
  )

  /** Mapping from app IDs to their app class */
  private[web] val appDirectory: mutable.Map[AppID, ServerApp] = mutable.Map()

  /** Mapping from app instance IDs to their running instances */
  private[web] val apps: concurrent.Map[InstanceID, RunningServerAppInstance] = concurrent.TrieMap()

  /** Mapping from app instance IDs to the attached running clocks */
  private[web] val clocks: concurrent.Map[InstanceID, RunningClock] = concurrent.TrieMap()

  private[web] lazy val webSocketServer: WebSocketsCollection = new WebSocketsCollection(Config.WS_PORT)

  /** Registers the given state-machine based app. */
  def register[E, V, S](sm: StateMachine[E, V, S]): Unit =
    register(StateMachineServerApp(sm))

  def isRegistered(smInstance: StateMachine[?, ?, ?]): Boolean =
    appDirectory.contains(smInstance.appInfo.id)

  /** Registers the given app */
  private def register(app: ServerApp): Unit =
    // Populate `appDirectory`
    val appID = app.appInfo.id
    this.appDirectory.put(appID, app)
    println(f"[$appID] registered")

  private[web] def createInstance(appID: AppID, clients: Seq[UserID]): InstanceID =
    // Populate `apps`
    def loadWordlist(name: String): Seq[String] =
      val fname = f"$name.txt"
      val wordstream =
        Option(this.getClass.getResourceAsStream(f"/$fname"))
          .getOrElse(sys.error(f"File not found: $fname"))
      try Source.fromInputStream(wordstream)(Codec.UTF8).getLines().toSeq
      finally wordstream.close()

    val words = loadWordlist(instanceIDsFileName)
    import scala.util.Random
    val instanceID = (Random.shuffle(words).take(instanceIDsCount) :+ Random.between(0, 1000).toString()).mkString("-")
    apps(instanceID) = RunningServerAppInstance(appDirectory(appID).init(clients), 0)

    apps(instanceID).instance match
      case StateMachineServerAppInstance(sm: ClockDrivenStateMachine[?, ?, ?], _, _) =>
        startClockFor(appID, instanceID, sm.clockPeriodMs)
      case StateMachineServerAppInstance(_, _, _) => ()

    webSocketServer.initializeApp(instanceID)
    println(f"[$appID] instance created $instanceID")
    instanceID

  private[web] def shutdownApp(instanceID: InstanceID): Unit =
    // Remove references of this app
    val appID = apps.remove(instanceID).map(_.instance.appInfo.id)
    clocks.get(instanceID).map(_._2).foreach(_.set(false))
    clocks.remove(instanceID)
    if appID.isDefined then
      println(s"[${appID.get}][$instanceID] shut down")

  private def startClockFor(appID: String, instanceID: InstanceID, clockResolutionMs: Int): Unit =
    object Clock extends Thread:
      override def run(): Unit =
        val clockName = s"clock-$instanceID"
        val running = AtomicBoolean(true)
        clocks.put(instanceID, RunningClock(clockName, running))
        setName(clockName)
        println(f"[$appID][$instanceID] clock started")
        while running.get() do
          handleMessage(
            instanceID,
            clockName,
            ujson.Obj(
              ClockDrivenStateMachine.CLOCK_EVENT_HEADER -> true,
              ClockDrivenStateMachine.CLOCK_EVENT_TICK_HEADER -> TickEventFormat.encode(Tick(System.currentTimeMillis))
            )
          )
          Thread.sleep(clockResolutionMs)

        println(s"[$appID][$instanceID] clock stopped")

    Clock.start()

  /** Creates a transition from the current state of the application to the next
    * state of the application.
    *
    * Given a [[Seq]] of [[UserID]], corresponding to the clients connected to
    * the app, accepts and event and the event's author's ID. The [[apps]] is
    * then used to handle the event and gives a sequence of actions for each
    * user.
    *
    * @param clients
    *   The set of clients currently connected in the application.
    * @param uid
    *   The user ID of the user who triggered the event.
    * @param event
    *   The event triggered by the user.
    * @return
    *   A map which links each user to a queue of views they should receive.
    */
  private def transition(clients: Seq[UserID], instanceID: InstanceID)(
      uID: UserID,
      event: ujson.Value
  ): Try[immutable.Map[UserID, Seq[Action[ujson.Value]]]] =
    val instance = apps(instanceID).instance
    instance.synchronized {
      instance.transition(clients)(uID, event).map {
        case (views, newApp) =>
          apps.update(instanceID, apps(instanceID).copy(instance = newApp))
          views
      }
    }

  /** Dispatches the given `msg`, sent by `userID` for app instance `instanceID`
    * appropriately to clients
    */
  private[web] def handleMessage(instanceID: InstanceID, userID: UserID, msg: ujson.Value): Unit =
    val transitionResult = transition(webSocketServer.connectedClients(instanceID), instanceID)(userID, msg)
    val serverResponse = transitionResult match
      case Failure(exception) =>
        // If sender is clock, send exception to all clients
        val senderIsClock = clocks.exists((clockID, _) => clockID == userID) // clocks use their ID as userID
        val recipients =
          if senderIsClock then webSocketServer.connectedClients(instanceID)
          else Seq(userID)

        for recipient <- recipients do
          webSocketServer.send(instanceID, recipient):
            EventResponse.Wire.encode(Failure(exception))
      case Success(userActions) =>
        for (uID, actions) <- userActions do
          webSocketServer.send(instanceID, uID):
            EventResponse.Wire.encode(Success(actions))
