package cs214.webapp
package server
package web

import java.net.InetAddress
import scala.jdk.CollectionConverters.*
import scala.util.Try
import cask.endpoints.JsonData

/** HTTP routes of the WebServer */
private[server] final case class WebServerRoutes()(using cc: castor.Context, log: cask.Logger) extends cask.Routes:
  /** Paths where the static content served by the server is stored */
  private val WEB_SRC_PATH = "www/static/"

  /** HTML page to serve when accessing the server `/` and `/app/...` path */
  private def HTML_STATIC_FILE =
    cask.model.StaticResource(
      WEB_SRC_PATH + "webapp.html",
      classOf[cask.staticResources].getClassLoader,
      Seq("Content-Type" -> "text/html")
    )

  lazy val hostAddress: String =
    val addresses =
      for
        intf <- java.net.NetworkInterface.getNetworkInterfaces.asScala
        if intf.isUp
        _ = if WebServer.debug then println(f"[debug] found interface $intf")
        addr <- intf.getInetAddresses.asScala
        if (addr.isInstanceOf[java.net.Inet4Address]
          && !addr.isLinkLocalAddress
          && !addr.isLoopbackAddress)
        _ = if WebServer.debug then println(f"[debug] found address ${addr.getHostAddress}")
      yield addr.getHostAddress
    Try(addresses.toList.head).getOrElse(InetAddress.getLocalHost.getHostAddress)

  @cask.get("/")
  def getIndexFile() = HTML_STATIC_FILE

  @cask.staticResources("/static")
  def getStaticResources() = WEB_SRC_PATH

  // For all /app subsegments, provide the HTML page
  @cask.get("/app")
  def getApp(segments: cask.RemainingPathSegments) = HTML_STATIC_FILE

  @cask.getJson(f"/${Endpoints.Api.listApps}")
  def getListApps() =
    ListAppsResponse.Wire.encode(ListAppsResponse(WebServer.appDirectory.values.map(_.appInfo).toSeq))

  @cask.getJson(f"/${Endpoints.Api.instanceInfo}")
  def getAppInfo(instanceID: InstanceID) =
    val response: cask.Response[JsonData] =
      if WebServer.apps.contains(instanceID) then
        val app = WebServer.apps(instanceID).instance
        val shareUrl = f"http://$hostAddress:${Config.HTTP_PORT}/app/${app.appInfo.id}/$instanceID/"
        val wsEndpoint = f"ws://{{hostName}}:${Config.WS_PORT}/$instanceID/{{userID}}"
        val response = InstanceInfoResponse(instanceID, app.registeredUsers, wsEndpoint, shareUrl)
        InstanceInfoResponse.Wire.encode(response)
      else
        cask.Response(f"Unknown instance ID $instanceID", 400)
    response

  @cask.post(f"/${Endpoints.Api.createInstance}")
  def postInitApp(request: cask.Request) =
    val response: cask.Response[JsonData] =
      val req = CreateInstanceRequest.Wire.decode(ujson.read(request.text()))
      if req.isFailure then
        cask.Response(f"Unable to decode data: ${request.text()}", 400)
      else
        val appID = WebServer.createInstance(req.get.appName, req.get.userIDs)
        CreateInstanceResponse.Wire.encode(CreateInstanceResponse(appID))
    response

  WebServer.webSocketServer.run()
  initialize()
