package cs214.webapp
package server.web

import org.java_websocket.WebSocket
import org.java_websocket.handshake.ClientHandshake
import org.java_websocket.server.WebSocketServer
import java.net.{InetSocketAddress, URLDecoder}
import scala.collection.mutable
import scala.util.{Success, Failure, Try}

/** A collection of websockets organized by app ID and client */
private[web] final class WebSocketsCollection(val port: Int):
  def onMessageReceive(appID: InstanceID, uID: UserID, msg: ujson.Value): Unit =
    WebServer.handleMessage(appID, uID, msg)

  // When the user joins the app, the projection of the current state is sent to them
  def onClientConnect(instanceID: InstanceID, userID: UserID): Unit =
    val instance = WebServer.apps(instanceID)
    val newInstance = instance.copy(connectedUsersCount = instance.connectedUsersCount + 1)
    WebServer.apps(instanceID) = newInstance
    val js = newInstance.instance.respondToNewClient(userID)
    send(instanceID, userID):
      EventResponse.Wire.encode(Success(List(Action.Render(js))))
    println(f"[${newInstance.instance.appInfo.id}][$instanceID] client \"$userID\" connected")

  def onClientDisconnect(instanceID: InstanceID, userID: UserID): Unit =
    val instance = WebServer.apps(instanceID)
    val newInstance = instance.copy(connectedUsersCount = instance.connectedUsersCount - 1)
    WebServer.apps(instanceID) = newInstance

    val appID = newInstance.instance.appInfo.id
    println(f"[$appID][$instanceID] client \"$userID\" disconnected")

  /** All the sessions currently in use, mapping instance IDs to an actual
    * socket object that a client owns
    */
  private val sessions: mutable.Map[InstanceID, Seq[(UserID, WebSocket)]] =
    mutable.Map()

  /** Initialize an empty session list for the given app instance */
  def initializeApp(instanceID: InstanceID) =
    require(!sessions.contains(instanceID))
    sessions(instanceID) = Seq.empty

  /** Runs k with parameters parsed from the websocket connection path. [[k]] is
    * run while synchronizing on [[sessions]]
    *
    * The connection should be on "ws://…/[app_instance_ID]/[user_ID]" for the
    * parsing to function properly.
    */
  private def withSessionParams[T](socket: WebSocket)(k: (InstanceID, UserID) => T) =
    sessions.synchronized:
      try
        val components = socket.getResourceDescriptor.split("/").takeRight(2)
        val decoded = components.map(s => URLDecoder.decode(s, "UTF-8"))
        decoded match
          case Array(appID, userID) =>
            if !sessions.contains(appID) then
              throw IllegalArgumentException("Error: Invalid app ID")
            k(appID, userID)
          case _ => throw Exception("Error: Invalid path")
      catch
        case t =>
          socket.send(SocketResponseWire.encode(util.Failure(t)).toString)
          socket.close()

  /** A single websocket server handling multiple apps and clients. */
  private val server: WebSocketServer = new WebSocketServer(InetSocketAddress("0.0.0.0", port)):
    override def onOpen(socket: WebSocket, handshake: ClientHandshake): Unit =
      withSessionParams(socket): (appID, clientID) =>
        sessions(appID) = sessions(appID) :+ (clientID, socket)
        onClientConnect(appID, clientID)

    override def onClose(socket: WebSocket, code: Int, reason: String, remote: Boolean): Unit =
      withSessionParams(socket): (appID, clientID) =>
        sessions(appID) = sessions(appID).filter(_._2 != socket) // Unregister the session
        onClientDisconnect(appID, clientID)

    override def onMessage(socket: WebSocket, message: String): Unit =
      withSessionParams(socket): (appID, clientID) =>
        onMessageReceive(appID, clientID, ujson.read(message))

    override def onError(socket: WebSocket, ex: Exception): Unit =
      // Only report the error, onClosed is called even when an error occurs
      throw new RuntimeException(ex)

    override def onStart(): Unit =
      val addr = server.getAddress
      println(s"[WebSocket] server started on ${addr.getHostName}:${addr.getPort}.")

  /** Starts the server asynchronously. */
  def run(): Unit =
    server.setReuseAddr(true) // Ignore leftover connections from pending processes
    Thread(() => server.run(), "Socket Thread").start()

  /** Enumerates clients connected to [[appID]]. */
  def connectedClients(appID: InstanceID): Seq[UserID] =
    sessions.get(appID).map(_.map(_._1).distinct).getOrElse(Seq())

  /** Sends a message to a specific client. */
  def send(appID: InstanceID, clientID: UserID)(message: ujson.Value) =
    val wrapped = SocketResponseWire.encode(util.Success(message)).toString
    sessions.synchronized:
      for
        (userID, socket) <- sessions.getOrElse(appID, Seq.empty)
        if userID == clientID
      do
        socket.send(wrapped)
