package cs214.webapp.utils

import java.net.URI
import java.net.http.HttpRequest
import scala.concurrent.duration.Duration
import scala.reflect.ClassTag
import cs214.webapp.*
import cs214.webapp.server.StateMachine

/** Abstract test suite to help testing logic of Webapps */
abstract class WebappSuite[Event, State, View] extends munit.FunSuite:

  protected val UID0: String = "yak"
  protected val UID1: String = "hut"
  protected val UID2: String = "kik"

  /** Mock user IDs that can be used in tests */
  protected val USER_IDS = Seq(UID0, UID1, UID2)

  protected val RNG = util.Random(0)

  /** Your app logic */
  val sm: StateMachine[Event, State, View]

  override val munitTimeout: Duration = Duration(10, "s")

  /// Testing helpers

  extension [U](u: U)
    /** Asserts that u is an instance of type V */
    inline def assertInstanceOf[V](using tag: ClassTag[V]): V =
      assert(u.isInstanceOf[V], "u has unexpected type "+u.getClass+"(expecting "+tag+" instead)")
      u.asInstanceOf[V]

  /** Asserts that this try is a success */
  def assertSuccess[V](ts: util.Try[V]): V =
    assert(ts.isSuccess, "Unexpected failure: "+ts)
    ts.get

  /** Asserts that actions contains a single action which is of type
    * Action.Render. In case of success, returns this render action.
    */
  def assertSingleRender[V](actions: Seq[Action[V]]): V =
    assertEquals(actions.length, 1)
    val r = actions.head.assertInstanceOf[Action.Render[V]]
    r.st

  /** Asserts that ts is a success and contains a single action which is of type
    * Action.Render. In case of success, returns this render action.
    */
  def assertSingleRender[V](ts: util.Try[Seq[Action[V]]]): V =
    assertSingleRender(assertSuccess(ts))

  /** Asserts that ts is a success and contains more than one action. In case of
    * success, returns these actions.
    */
  def assertMultipleActions[K](ts: util.Try[Seq[Action[K]]], n: Int): Seq[Action[K]] =
    val actions = assertSuccess(ts)
    assertEquals(actions.length, n)
    actions.map(_.asInstanceOf[Action[K]])

  // [Any] is a trick to avoid using [?]
  /** Asserts that ts is a failure and returns this failure. */
  inline def assertFailure[F <: Throwable](ts: util.Try[Seq[Action[Any]]])(using ClassTag[F]): F =
    assert(ts.isFailure)
    ts.failed.get.assertInstanceOf[F]

  /** Asserts that encoding then decoding t succeeds and doesn't modify t. In
    * case of success, returns t.
    */
  def testWire[T](w: WireFormat[T])(t: T): T =
    val d = w.decode(w.encode(t))
    assert(d.isSuccess)
    assertEquals(d.get, t)
    t

  extension (v: View)
    def testViewWire: View = testWire(sm.wire.viewFormat)(v)

  extension (e: Event)
    def testEventWire: Event = testWire(sm.wire.eventFormat)(e)
